@@ -2,11 +2,11 @@
 /*                                                                          */
 /*  The FreeType project -- a free and portable quality TrueType renderer.  */
 /*                                                                          */
-/*  Copyright 1996-2000, 2003-2007, 2009-2013 by                            */
+/*  Copyright 1996-2000, 2003-2007, 2009-2014 by                            */
 /*  D. Turner, R.Wilhelm, and W. Lemberg                                    */
 /*                                                                          */
 /*                                                                          */
-/*  FTGrid - a simple viewer to debug the auto-hinter                       */
+/*  FTGrid - a simple viewer to show glyph outlines on a grid               */
 /*                                                                          */
 /*  Press F1 when running this program to have a list of key-bindings       */
 /*                                                                          */
@@ -19,8 +19,16 @@
 
   /* the following header shouldn't be used in normal programs */
 #include FT_INTERNAL_DEBUG_H
+
+  /* showing driver name */
+#include FT_MODULE_H
+#include FT_INTERNAL_OBJECTS_H
+#include FT_INTERNAL_DRIVER_H
+
 #include FT_STROKER_H
 #include FT_SYNTHESIS_H
+#include FT_CFF_DRIVER_H
+#include FT_TRUETYPE_DRIVER_H
 
 #define MAXPTSIZE      500                 /* dtp */
 
@@ -36,27 +44,33 @@
 #define snprintf  _snprintf
 #endif
 
+#define N_CFF_HINTING_ENGINES  2
 
+
+#ifdef FT_DEBUG_AUTOFIT
   /* these variables, structures and declarations are for   */
   /* communication with the debugger in the autofit module; */
   /* normal programs don't need this                        */
   struct  AF_GlyphHintsRec_;
   typedef struct AF_GlyphHintsRec_*  AF_GlyphHints;
 
-  int            _af_debug_disable_horz_hints;
-  int            _af_debug_disable_vert_hints;
-  int            _af_debug_disable_blue_hints;
-  AF_GlyphHints  _af_debug_hints;
+  extern int            _af_debug_disable_horz_hints;
+  extern int            _af_debug_disable_vert_hints;
+  extern int            _af_debug_disable_blue_hints;
+  extern AF_GlyphHints  _af_debug_hints;
 
 #ifdef __cplusplus
   extern "C" {
 #endif
   extern void
-  af_glyph_hints_dump_segments( AF_GlyphHints  hints );
+  af_glyph_hints_dump_segments( AF_GlyphHints  hints,
+                                FT_Bool        to_stdout );
   extern void
-  af_glyph_hints_dump_points( AF_GlyphHints  hints );
+  af_glyph_hints_dump_points( AF_GlyphHints  hints,
+                              FT_Bool        to_stdout );
   extern void
-  af_glyph_hints_dump_edges( AF_GlyphHints  hints );
+  af_glyph_hints_dump_edges( AF_GlyphHints  hints,
+                             FT_Bool        to_stdout );
   extern FT_Error
   af_glyph_hints_get_num_segments( AF_GlyphHints  hints,
                                    FT_Int         dimension,
@@ -65,11 +79,15 @@
   af_glyph_hints_get_segment_offset( AF_GlyphHints  hints,
                                      FT_Int         dimension,
                                      FT_Int         idx,
-                                     FT_Pos*        offset );
+                                     FT_Pos        *offset,
+                                     FT_Bool       *is_blue,
+                                     FT_Pos        *blue_offset );
 #ifdef __cplusplus
   }
 #endif
 
+#endif /* FT_DEBUG_AUTOFIT */
+
   typedef struct  GridStatusRec_
   {
     int          width;
@@ -97,9 +115,9 @@
     grColor      grid_color;
     grColor      outline_color;
     grColor      on_color;
-    grColor      conic_color;
-    grColor      cubic_color;
+    grColor      off_color;
     grColor      segment_color;
+    grColor      blue_color;
 
     int          do_horz_hints;
     int          do_vert_hints;
@@ -112,6 +130,9 @@
     const char*  header;
     char         header_buffer[256];
 
+    int          cff_hinting_engine;
+    int          tt_interpreter_version;
+
   } GridStatusRec, *GridStatus;
 
   static GridStatusRec  status;
@@ -149,13 +170,13 @@
     st->disp_height   = display->bitmap->rows;
     st->disp_bitmap   = display->bitmap;
 
-    st->axis_color    = grFindColor( display->bitmap,   0,   0,   0, 255 );
-    st->grid_color    = grFindColor( display->bitmap, 192, 192, 192, 255 );
-    st->outline_color = grFindColor( display->bitmap, 255,   0,   0, 255 );
-    st->on_color      = grFindColor( display->bitmap,  64,  64, 255, 255 );
-    st->conic_color   = grFindColor( display->bitmap,   0, 128,   0, 255 );
-    st->cubic_color   = grFindColor( display->bitmap, 255,  64, 255, 255 );
-    st->segment_color = grFindColor( display->bitmap,  64, 255, 128,  64 );
+    st->axis_color    = grFindColor( display->bitmap,   0,   0,   0, 255 ); /* black       */
+    st->grid_color    = grFindColor( display->bitmap, 192, 192, 192, 255 ); /* gray        */
+    st->outline_color = grFindColor( display->bitmap, 255,   0,   0, 255 ); /* red         */
+    st->on_color      = grFindColor( display->bitmap, 255,   0,   0, 255 ); /* red         */
+    st->off_color     = grFindColor( display->bitmap,   0, 128,   0, 255 ); /* dark green  */
+    st->segment_color = grFindColor( display->bitmap,  64, 255, 128,  64 ); /* light green */
+    st->blue_color    = grFindColor( display->bitmap,  64,  64, 255,  64 ); /* light blue  */
   }
 
 
@@ -184,12 +205,15 @@
       xmax  = ( xmax + 63 ) & ~63;
       ymax  = ( ymax + 63 ) & ~63;
 
-      printf( "XXX x_ppem=%d y_ppem=%d width=%d height=%d\n",
-              size->metrics.x_ppem, size->metrics.y_ppem,
-              xmax - xmin, ymax - ymin );
+      if ( xmax - xmin )
+        x_scale = st->disp_width  * ( 1.0 - 2 * st->margin ) / ( xmax - xmin );
+      else
+        x_scale = 1.0;
 
-      x_scale = st->disp_width  * ( 1.0 - 2 * st->margin ) / ( xmax - xmin );
-      y_scale = st->disp_height * ( 1.0 - 2 * st->margin ) / ( ymax - ymin );
+      if ( ymax - ymin )
+        y_scale = st->disp_height * ( 1.0 - 2 * st->margin ) / ( ymax - ymin );
+      else
+        y_scale = 1.0;
 
       if ( x_scale <= y_scale )
         st->scale = x_scale;
@@ -250,6 +274,8 @@
   }
 
 
+#ifdef FT_DEBUG_AUTOFIT
+
   static void
   grid_hint_draw_segment( GridStatus     st,
                           AF_GlyphHints  hints )
@@ -269,12 +295,15 @@
 
       for ( count = 0; count < num_seg; count++ )
       {
-        int     pos;
-        FT_Pos  offset;
+        int      pos;
+        FT_Pos   offset;
+        FT_Bool  is_blue;
+        FT_Pos   blue_offset;
 
 
         af_glyph_hints_get_segment_offset( hints, dimension,
-                                           count, &offset );
+                                           count, &offset,
+                                           &is_blue, &blue_offset);
 
         if ( dimension == 0 ) /* AF_DIMENSION_HORZ is 0 */
         {
@@ -285,13 +314,33 @@
         else
         {
           pos = y_org - (int)offset * st->scale;
-          grFillHLine( st->disp_bitmap, 0, pos,
-                       st->disp_width, st->segment_color );
+
+          if ( is_blue )
+          {
+            int  blue_pos = y_org - (int)blue_offset * st->scale;
+
+
+            if ( blue_pos == pos )
+              grFillHLine( st->disp_bitmap, 0, blue_pos,
+                           st->disp_width, st->blue_color );
+            else
+            {
+              grFillHLine( st->disp_bitmap, 0, blue_pos,
+                           st->disp_width, st->blue_color );
+              grFillHLine( st->disp_bitmap, 0, pos,
+                           st->disp_width, st->segment_color );
+            }
+          }
+          else
+            grFillHLine( st->disp_bitmap, 0, pos,
+                         st->disp_width, st->segment_color );
         }
       }
     }
   }
 
+#endif /* FT_DEBUG_AUTOFIT */
+
 
   static void
   ft_bitmap_draw( FT_Bitmap*       bitmap,
@@ -478,6 +527,7 @@
 
     FTDemo_Get_Size( handle, &size );
 
+#ifdef FT_DEBUG_AUTOFIT
     /* Draw segment before drawing glyph. */
     if ( status.do_segment )
     {
@@ -495,6 +545,7 @@
 
     _af_debug_disable_horz_hints = !st->do_horz_hints;
     _af_debug_disable_vert_hints = !st->do_vert_hints;
+#endif
 
     if ( FT_Load_Glyph( size->face, st->Num,
                         handle->load_flags | FT_LOAD_NO_BITMAP ) )
@@ -541,7 +592,7 @@
             handle,
             display,
             ( gimage->tags[nn] & FT_CURVE_TAG_ON ) ? st->on_color
-                                                   : st->conic_color );
+                                                   : st->off_color );
       }
     }
   }
@@ -588,56 +639,42 @@
     grSetMargin( 2, 1 );
     grGotobitmap( display->bitmap );
 
-    grWriteln( "FreeType Glyph Viewer - part of the FreeType test suite" );
-    grLn();
-    grWriteln( "This program is used to display all glyphs from one or" );
-    grWriteln( "several font files, with the FreeType library." );
+    grWriteln( "FreeType Glyph Grid Viewer - part of the FreeType test suite" );
     grLn();
     grWriteln( "Use the following keys:" );
     grLn();
-    grWriteln( "  F1 or ?    : display this help screen" );
-    grLn();
-    grWriteln( "  a          : toggle anti-aliasing" );
-    grWriteln( "  f          : toggle forced autofit mode" );
-    grWriteln( "  h          : toggle ignore hinting mode" );
-    grWriteln( "  left/right : decrement/increment glyph index" );
-    grWriteln( "  up/down    : change character size" );
-    grLn();
-    grWriteln( "  F7         : decrement index by 10" );
-    grWriteln( "  F8         : increment index by 10" );
-    grWriteln( "  F9         : decrement index by 100" );
-    grWriteln( "  F10        : increment index by 100" );
-    grWriteln( "  F11        : decrement index by 1000" );
-    grWriteln( "  F12        : increment index by 1000" );
-    grLn();
-    grWriteln( "  i          : move grid up" );
-    grWriteln( "  j          : move grid left" );
-    grWriteln( "  k          : move grid down" );
-    grWriteln( "  l          : move grid right" );
-    grWriteln( "  Page up/dn : zoom in/out grid" );
-    grWriteln( "  RETURN     : reset zoom and position" );
-    grLn();
+    /*          |----------------------------------|    |----------------------------------| */
 #ifdef FT_DEBUG_AUTOFIT
-    grWriteln( "  H          : toggle horizontal hinting" );
-    grWriteln( "  V          : toggle vertical hinting" );
-    grWriteln( "  B          : toggle blue zone hinting" );
-    grWriteln( "  s          : toggle segment drawing" );
-#endif
-    grWriteln( "  d          : toggle dots display" );
-    grWriteln( "  o          : toggle outline display" );
-    grWriteln( "  g          : increase gamma by 0.1" );
-    grWriteln( "  v          : decrease gamma by 0.1" );
+    grWriteln( "F1, ?       display this help screen    if autohinting:                     " );
+    grWriteln( "                                          H         toggle horiz. hinting   " );
+    grWriteln( "i, k        move grid up/down             V         toggle vert. hinting    " );
+    grWriteln( "j, l        move grid left/right          B         toggle blue zone hinting" );
+    grWriteln( "PgUp, PgDn  zoom in/out grid              s         toggle segment drawing  " );
+    grWriteln( "SPC         reset zoom and position                 (unfitted, with blues)  " );
+    grWriteln( "                                          1         dump edge hints         " );
+    grWriteln( "p, n        previous/next font            2         dump segment hints      " );
+    grWriteln( "                                          3         dump point hints        " );
+#else
+    grWriteln( "F1, ?       display this help screen    i, k        move grid up/down       " );
+    grWriteln( "                                        j, l        move grid left/right    " );
+    grWriteln( "p, n        previous/next font          PgUp, PgDn  zoom in/out grid        " );
+    grWriteln( "                                        SPC         reset zoom and position " );
+#endif /* FT_DEBUG_AUTOFIT */
+    grWriteln( "Up, Down    adjust size by 0.5pt                                            " );
+    grWriteln( "                                        if not autohinting:                 " );
+    grWriteln( "Left, Right adjust index by 1             H         cycle through hinting   " );
+    grWriteln( "F7, F8      adjust index by 10                        engines (if available)" );
+    grWriteln( "F9, F10     adjust index by 100                                             " );
+    grWriteln( "F11, F12    adjust index by 1000        d           toggle dots display     " );
+    grWriteln( "                                        o           toggle outline display  " );
+    grWriteln( "h           toggle hinting                                                  " );
+    grWriteln( "f           toggle forced auto-                                             " );
+    grWriteln( "             hinting (if hinting)       g, v        adjust gamma value      " );
+    grWriteln( "                                                                            " );
+    grWriteln( "a           toggle anti-aliasing        q, ESC      quit ftgrid             " );
+    /*          |----------------------------------|    |----------------------------------| */
     grLn();
-    grWriteln( "  n          : next font" );
-    grWriteln( "  p          : previous font" );
-    grWriteln( "  q / ESC    : exit program" );
     grLn();
-#ifdef FT_DEBUG_AUTOFIT
-    grWriteln( "  1          : dump edge hints" );
-    grWriteln( "  2          : dump segment hints" );
-    grWriteln( "  3          : dump point hints" );
-    grLn();
-#endif
     grWriteln( "press any key to exit this help screen" );
 
     grRefreshSurface( display->surface );
@@ -645,6 +682,73 @@
   }
 
 
+  static void
+  event_cff_hinting_engine_change( int  delta )
+  {
+    int  new_cff_hinting_engine;
+
+
+    if ( delta )
+      new_cff_hinting_engine =
+        ( status.cff_hinting_engine +
+          delta                     +
+          N_CFF_HINTING_ENGINES     ) % N_CFF_HINTING_ENGINES;
+
+    error = FT_Property_Set( handle->library,
+                             "cff",
+                             "hinting-engine",
+                             &new_cff_hinting_engine );
+
+    if ( !error )
+    {
+      /* Resetting the cache is perhaps a bit harsh, but I'm too  */
+      /* lazy to walk over all loaded fonts to check whether they */
+      /* are of type CFF, then unloading them explicitly.         */
+      FTC_Manager_Reset( handle->cache_manager );
+      status.cff_hinting_engine = new_cff_hinting_engine;
+    }
+
+    sprintf( status.header_buffer, "CFF engine changed to %s",
+             status.cff_hinting_engine == FT_CFF_HINTING_FREETYPE
+               ? "FreeType" : "Adobe" );
+
+    status.header = (const char *)status.header_buffer;
+  }
+
+
+  static void
+  event_tt_interpreter_version_change( void )
+  {
+    FT_UInt  new_interpreter_version;
+
+
+    if ( status.tt_interpreter_version == TT_INTERPRETER_VERSION_35 )
+      new_interpreter_version = TT_INTERPRETER_VERSION_38;
+    else
+      new_interpreter_version = TT_INTERPRETER_VERSION_35;
+
+    error = FT_Property_Set( handle->library,
+                             "truetype",
+                             "interpreter-version",
+                             &new_interpreter_version );
+
+    if ( !error )
+    {
+      /* Resetting the cache is perhaps a bit harsh, but I'm too  */
+      /* lazy to walk over all loaded fonts to check whether they */
+      /* are of type TTF, then unloading them explicitly.         */
+      FTC_Manager_Reset( handle->cache_manager );
+      status.tt_interpreter_version = new_interpreter_version;
+    }
+
+    sprintf( status.header_buffer, "TrueType engine changed to version %s",
+             status.tt_interpreter_version == TT_INTERPRETER_VERSION_35
+               ? "35" : "38" );
+
+    status.header = (const char *)status.header_buffer;
+  }
+
+
   static void
   event_gamma_change( double  delta )
   {
@@ -787,30 +891,24 @@
       if ( handle->hinted && handle->autohint )
       {
         status.header = "dumping glyph edges to stdout";
-        af_glyph_hints_dump_edges( _af_debug_hints );
+        af_glyph_hints_dump_edges( _af_debug_hints, 1 );
       }
-      else
-        status.header = "need autofit mode for edge dumping";
       break;
 
     case grKEY( '2' ):
       if ( handle->hinted && handle->autohint )
       {
         status.header = "dumping glyph segments to stdout";
-        af_glyph_hints_dump_segments( _af_debug_hints );
+        af_glyph_hints_dump_segments( _af_debug_hints, 1 );
       }
-      else
-        status.header = "need autofit mode for segment dumping";
       break;
 
     case grKEY( '3' ):
       if ( handle->hinted && handle->autohint )
       {
         status.header = "dumping glyph points to stdout";
-        af_glyph_hints_dump_points( _af_debug_hints );
+        af_glyph_hints_dump_points( _af_debug_hints, 1 );
       }
-      else
-        status.header = "need autofit mode for point dumping";
       break;
 #endif /* FT_DEBUG_AUTOFIT */
 
@@ -846,18 +944,36 @@
       event_font_change( -1 );
       break;
 
-#ifdef FT_DEBUG_AUTOFIT
     case grKEY( 'H' ):
-      if ( handle->autohint )
+      if ( !handle->autohint )
+      {
+        FT_Face    face;
+        FT_Module  module;
+
+
+        error = FTC_Manager_LookupFace( handle->cache_manager,
+                                        handle->scaler.face_id, &face );
+        if ( !error )
+        {
+          module = &face->driver->root;
+
+          if ( !strcmp( module->clazz->module_name, "cff" ) )
+            event_cff_hinting_engine_change( 1 );
+          else if ( !strcmp( module->clazz->module_name, "truetype" ) )
+            event_tt_interpreter_version_change();
+        }
+      }
+#ifdef FT_DEBUG_AUTOFIT
+      else
       {
         status.do_horz_hints = !status.do_horz_hints;
         status.header = status.do_horz_hints ? "horizontal hinting enabled"
                                              : "horizontal hinting disabled";
       }
-      else
-        status.header = "need autofit mode to toggle horizontal hinting";
+#endif
       break;
 
+#ifdef FT_DEBUG_AUTOFIT
     case grKEY( 'V' ):
       if ( handle->autohint )
       {
@@ -1134,6 +1250,14 @@
     grid_status_init( &status );
     parse_cmdline( &argc, &argv );
 
+    /* get the default value as compiled into FreeType */
+    FT_Property_Get( handle->library,
+                     "cff",
+                     "hinting-engine", &status.cff_hinting_engine );
+    FT_Property_Get( handle->library,
+                     "truetype",
+                     "interpreter-version", &status.tt_interpreter_version );
+
     display = FTDemo_Display_New( gr_pixel_mode_rgb24,
                                   status.width, status.height );
     if ( !display )
@@ -1145,7 +1269,12 @@
                 "FreeType Glyph Grid Viewer - press F1 for help" );
 
     for ( ; argc > 0; argc--, argv++ )
-      FTDemo_Install_Font( handle, argv[0] );
+    {
+      error = FTDemo_Install_Font( handle, argv[0], 1 );
+      if ( error == FT_Err_Invalid_Argument )
+        fprintf( stderr, "skipping font `%s' without outlines\n",
+                         argv[0] );
+    }
 
     if ( handle->num_fonts == 0 )
       Fatal( "could not find/open any font file" );
